import os
from pathlib import Path
from typing import List

import pytest

from rakam_systems_tools.evaluation.client import DeepEvalClient
from rakam_systems_cli.decorators import eval_run
from rakam_systems_cli.utils.decorator_utils import find_decorated_functions


@pytest.fixture
def client() -> DeepEvalClient:
    return DeepEvalClient(base_url="http://testserver", api_token="testtoken")


class FakeCPUTimes:
    def __init__(self, user: float = 1.0, system: float = 1.0) -> None:
        self.user: float = user
        self.system: float = system


class FakeMemInfo:
    def __init__(self, rss: int) -> None:
        self.rss: int = rss


class FakeProcess:
    def __init__(self) -> None:
        self._cpu_calls: int = 0
        self._mem_calls: int = 0

    def cpu_times(self) -> FakeCPUTimes:
        self._cpu_calls += 1
        # simulate CPU usage increase
        return FakeCPUTimes(
            user=1.0 + self._cpu_calls,
            system=1.0,
        )

    def memory_info(self) -> FakeMemInfo:
        self._mem_calls += 1
        return FakeMemInfo(rss=100_000_000 + (self._mem_calls * 10_000))


@pytest.fixture(autouse=True)
def patch_psutil(
    monkeypatch: pytest.MonkeyPatch,
) -> FakeProcess:
    fake_process: FakeProcess = FakeProcess()

    monkeypatch.setattr(
        "rakam_systems_cli.decorators.psutil.Process",
        lambda pid: fake_process,
    )
    monkeypatch.setattr(os, "getpid", lambda: 123)
    return fake_process


def test_eval_run_basic(capsys: pytest.CaptureFixture[str]) -> None:
    @eval_run
    def add(a: int, b: int) -> int:
        return a + b

    result: int = add(2, 3)

    assert result == 5

    out: str = capsys.readouterr().out
    assert "[eval_run]" in out
    assert "add" in out
    assert "time=" in out
    assert "cpu=" in out
    assert "mem_delta=" in out


def test_eval_run_with_parentheses(capsys: pytest.CaptureFixture[str]) -> None:
    @eval_run()
    def mul(a: int, b: int) -> int:
        return a * b

    result: int = mul(3, 4)

    assert result == 12
    assert "[eval_run]" in capsys.readouterr().out


def test_find_decorated_functions(tmp_path: Path) -> None:
    code = """
from rakam_systems_cli.decorators import eval_run

@eval_run
def foo():
    pass

@eval_run()
def bar():
    pass

async def baz():
    pass

@other
def nope():
    pass
"""
    file = tmp_path / "test_mod.py"
    file.write_text(code)

    result: List[str] = find_decorated_functions(file, "eval_run")

    assert set(result) == {"foo", "bar"}



import json
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Tuple

import pytest
import typer

from rakam_systems_cli.utils.print import (
    _print_and_save,
    git_diff,
    metric_direction,
    pct_change,
    pretty_print_comparison,
    print_metric_diff,
    print_summary,
    serialize_for_diff,
    summarize,
)


@pytest.mark.parametrize(
    "a,b,expected",
    [
        (10, 20, "+100.00%"),
        (20, 10, "-50.00%"),
        (10, 10, "+0.00%"),
        (None, 10, None),
        (10, None, None),
        (0, 10, None),
    ],
)
def test_pct_change(
    a: Optional[int], b: Optional[int], expected: Optional[str]
) -> None:
    assert pct_change(a, b) == expected


# -------------------------
# metric_direction
# -------------------------


@pytest.mark.parametrize(
    "delta,expected",
    [
        (1.0, "improved"),
        (-1.0, "regressed"),
        (0.0, "unchanged"),
        (None, "unchanged"),
    ],
)
def test_metric_direction(delta: Optional[float], expected: str) -> None:
    assert metric_direction(delta) == expected


def test_print_and_save_no_output_path(capsys: pytest.CaptureFixture[str]) -> None:
    resp: Dict[str, int] = {"a": 1}

    _print_and_save(resp, pretty=False, out=None, overwrite=False)

    captured = capsys.readouterr()
    assert str(resp) in captured.out


def test_print_and_save_creates_file(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    out: Path = tmp_path / "result.json"
    resp: Dict[str, str] = {"hello": "world"}

    _print_and_save(resp, pretty=False, out=out, overwrite=False)

    assert out.exists()
    assert json.loads(out.read_text()) == resp

    captured = capsys.readouterr()
    assert "Result saved" in captured.out


def test_print_and_save_refuses_overwrite(tmp_path: Path) -> None:
    out: Path = tmp_path / "result.json"
    out.write_text("{}")

    with pytest.raises(typer.Exit):
        _print_and_save({"x": 1}, pretty=False, out=out, overwrite=False)


def test_print_and_save_overwrites(tmp_path: Path) -> None:
    out: Path = tmp_path / "result.json"
    out.write_text("{}")

    _print_and_save({"x": 2}, pretty=False, out=out, overwrite=True)

    assert json.loads(out.read_text()) == {"x": 2}


def make_metric(
    *,
    metric: str,
    status: str = "changed",
    delta: Optional[float] = None,
) -> SimpleNamespace:
    return SimpleNamespace(
        metric=metric,
        status=status,
        delta=delta,
    )


def test_summarize_groups_all_states() -> None:
    metrics: List[SimpleNamespace] = [
        make_metric(metric="added_metric", status="added"),
        make_metric(metric="removed_metric", status="removed"),
        make_metric(metric="improved_metric", delta=1.0),
        make_metric(metric="regressed_metric", delta=-1.0),
        make_metric(metric="unchanged_zero", delta=0.0),
        make_metric(metric="unchanged_none", delta=None),
    ]

    summary: Dict[str, List[str]] = summarize(
        metrics=metrics)  # type: ignore[arg-type]

    assert summary == {
        "improved": ["improved_metric"],
        "regressed": ["regressed_metric"],
        "unchanged": ["unchanged_zero", "unchanged_none"],
        "added": ["added_metric"],
        "removed": ["removed_metric"],
    }


def test_summarize_empty_metrics() -> None:
    assert summarize(metrics=[]) == {
        "improved": [],
        "regressed": [],
        "unchanged": [],
        "added": [],
        "removed": [],
    }


def test_pretty_print_comparison_summary_only(monkeypatch: pytest.MonkeyPatch) -> None:
    metrics: List[SimpleNamespace] = [
        make_metric(metric="added_metric", status="added"),
        make_metric(metric="improved_metric", delta=1.0),
    ]
    resp: SimpleNamespace = SimpleNamespace(metrics=metrics)

    called: Dict[str, Any] = {}

    def fake_print_summary(arg: Any) -> None:
        called["metrics"] = arg

    import rakam_systems_cli.utils.print as mod

    monkeypatch.setattr(mod, "print_summary", fake_print_summary)

    pretty_print_comparison(resp, summary_only=True)

    assert called["metrics"] == metrics


def test_pretty_print_comparison_full(monkeypatch: pytest.MonkeyPatch) -> None:
    metrics: List[SimpleNamespace] = [
        make_metric(metric="added_metric", status="added"),
        make_metric(metric="improved_metric", delta=1.0),
    ]
    resp: SimpleNamespace = SimpleNamespace(metrics=metrics)

    calls: List[Any] = []

    def fake_print_metric_diff(metric: Any) -> None:
        calls.append(metric)

    import rakam_systems_cli.utils.print as mod

    monkeypatch.setattr(mod, "print_metric_diff", fake_print_metric_diff)

    pretty_print_comparison(resp, summary_only=False)

    assert calls == metrics


# Helper to capture secho calls
def capture_secho(monkeypatch: pytest.MonkeyPatch) -> List[Tuple[str, dict]]:
    calls: List[Tuple[str, dict]] = []

    def fake_secho(message: str, **kwargs: Any) -> None:
        calls.append((message, kwargs))

    monkeypatch.setattr(
        "rakam_systems_cli.utils.print.secho",
        fake_secho,
    )
    return calls


def test_print_metric_diff_added(monkeypatch: pytest.MonkeyPatch) -> None:
    calls = capture_secho(monkeypatch)

    diff = SimpleNamespace(
        metric="accuracy",
        status="added",
        score_b=0.9,
        threshold_b=0.8,
        success_b=True,
    )

    print_metric_diff(diff)  # type: ignore[arg-type]

    messages = [msg for msg, _ in calls]

    assert any("Metric: accuracy" in m for m in messages)
    assert any("+ score: 0.9" in m for m in messages)
    assert any("+ threshold: 0.8" in m for m in messages)
    assert any("+ success: True" in m for m in messages)


def test_print_metric_diff_removed(monkeypatch: pytest.MonkeyPatch) -> None:
    calls = capture_secho(monkeypatch)

    diff = SimpleNamespace(
        metric="latency",
        status="removed",
        score_a=120,
        threshold_a=100,
        success_a=False,
    )

    print_metric_diff(diff)  # type: ignore[arg-type]

    messages = [msg for msg, _ in calls]

    assert any("Metric: latency" in m for m in messages)
    assert any("- score: 120" in m for m in messages)
    assert any("- threshold: 100" in m for m in messages)
    assert any("- success: False" in m for m in messages)


def test_print_metric_diff_changed_improved(monkeypatch: pytest.MonkeyPatch) -> None:
    calls = capture_secho(monkeypatch)

    diff = SimpleNamespace(
        metric="f1",
        status="changed",
        score_a=0.5,
        score_b=0.8,
        threshold_a=0.7,
        threshold_b=0.7,
        success_a=False,
        success_b=True,
        delta=0.3,
    )

    print_metric_diff(diff)  # type: ignore[arg-type]

    messages = [msg for msg, _ in calls]

    assert any("- score: 0.5" in m for m in messages)
    assert any("+ score: 0.8" in m for m in messages)
    assert any("(+60.00%)" in m for m in messages)
    assert any("threshold: 0.7" in m for m in messages)
    assert any("- success: False" in m for m in messages)
    assert any("+ success: True" in m for m in messages)


def test_print_metric_diff_unchanged(monkeypatch: pytest.MonkeyPatch) -> None:
    calls = capture_secho(monkeypatch)

    diff = SimpleNamespace(
        metric="precision",
        status="unchanged",
        score_a=0.9,
        score_b=0.9,
        threshold_a=0.8,
        threshold_b=0.8,
        success_a=True,
        success_b=True,
        delta=0.0,
    )

    print_metric_diff(diff)  # type: ignore[arg-type]

    messages = [msg for msg, _ in calls]

    assert any("score: 0.9" in m for m in messages)
    assert any("threshold: 0.8" in m for m in messages)
    assert any("success: True" in m for m in messages)


def test_serialize_for_diff_stable_sorted_output() -> None:
    obj = {
        "b": 1,
        "a": 2,
        "nested": {
            "z": 3,
            "y": 4,
        },
    }

    result: str = serialize_for_diff(obj)

    # must be valid JSON
    parsed = json.loads(result)
    assert parsed == obj

    # sorted keys (a before b)
    assert result.index('"a"') < result.index('"b"')

    # newline at EOF (git-friendly)
    assert result.endswith("\n")


def test_serialize_for_diff_unicode_preserved() -> None:
    obj = {"text": "café 🚀"}

    result: str = serialize_for_diff(obj)

    assert "café" in result
    assert "🚀" in result
    assert "\\u" not in result


def test_git_diff_side_by_side_requires_git(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: List[str] = []

    monkeypatch.setattr("shutil.which", lambda _: None)
    monkeypatch.setattr("typer.secho", lambda msg, **_: calls.append(msg))

    git_diff(
        "a",
        "b",
        label_a="a.txt",
        label_b="b.txt",
        side_by_side=True,
    )

    assert any("Git is required" in msg for msg in calls)


def test_git_diff_side_by_side_requires_vimdiff(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: List[str] = []

    def fake_which(cmd: str) -> Optional[str]:
        return "/usr/bin/git" if cmd == "git" else None

    monkeypatch.setattr("shutil.which", fake_which)
    monkeypatch.setattr("typer.secho", lambda msg, **_: calls.append(msg))

    git_diff(
        "a",
        "b",
        label_a="a.txt",
        label_b="b.txt",
        side_by_side=True,
    )

    assert any("Vimdiff is not installed" in msg for msg in calls)


def test_git_diff_side_by_side_runs_difftool(monkeypatch: pytest.MonkeyPatch) -> None:
    run_calls: List[List[str]] = []

    monkeypatch.setattr("shutil.which", lambda _: "/usr/bin/tool")
    monkeypatch.setattr(
        "subprocess.run",
        lambda cmd, check=False: run_calls.append(cmd),
    )

    git_diff(
        "foo",
        "bar",
        label_a="a.txt",
        label_b="b.txt",
        side_by_side=True,
    )

    assert run_calls
    assert "difftool" in run_calls[0]
    assert "--tool=vimdiff" in run_calls[0]


def test_git_diff_uses_git_diff_when_available(monkeypatch: pytest.MonkeyPatch) -> None:
    run_calls: List[List[str]] = []

    def fake_which(cmd: str) -> Optional[str]:
        return "/usr/bin/git" if cmd == "git" else None

    monkeypatch.setattr("shutil.which", fake_which)
    monkeypatch.setattr(
        "subprocess.run",
        lambda cmd, check=False: run_calls.append(cmd),
    )

    git_diff(
        "a",
        "b",
        label_a="old.txt",
        label_b="new.txt",
        side_by_side=False,
    )

    assert run_calls
    assert run_calls[0][1] == "diff"
    assert "--no-index" in run_calls[0]
    assert "--color=always" in run_calls[0]


def test_git_diff_fallback_to_difflib(monkeypatch: pytest.MonkeyPatch) -> None:
    output: List[str] = []

    monkeypatch.setattr("shutil.which", lambda _: None)
    monkeypatch.setattr("typer.echo", lambda line: output.append(line))

    git_diff(
        "line1\nline2",
        "line1\nline3",
        label_a="a.txt",
        label_b="b.txt",
        side_by_side=False,
    )

    assert any("--- a.txt" in line for line in output)
    assert any("+++ b.txt" in line for line in output)
    assert any("-line2" in line for line in output)
    assert any("+line3" in line for line in output)


# Adjust this import to your actual module


@pytest.fixture
def fake_metrics() -> List[SimpleNamespace]:
    return [
        SimpleNamespace(name="accuracy"),
        SimpleNamespace(name="f1"),
    ]


@pytest.fixture
def summary_result() -> Dict:
    return {
        "improved": ["accuracy"],
        "regressed": ["latency"],
        "unchanged": [],
        "added": ["precision", "recall"],
        "removed": ["auc"],
    }


def test_print_summary_outputs_expected_lines(
    monkeypatch: pytest.MonkeyPatch,
    fake_metrics: List,
    summary_result: Dict,
) -> None:
    secho_calls = []

    def fake_secho(*args: Tuple, **kwargs: Dict) -> None:
        secho_calls.append((args, kwargs))

    monkeypatch.setattr(
        "rakam_systems_cli.utils.print.summarize",
        lambda _: summary_result,
    )
    monkeypatch.setattr(
        "rakam_systems_cli.utils.print._fmt",
        lambda metrics: ", ".join(metrics) if metrics else "-",
    )
    monkeypatch.setattr(
        "rakam_systems_cli.utils.print.secho",
        fake_secho,
    )

    print_summary(fake_metrics)

    # --- Header ---
    assert secho_calls[0][0][0] == "\nSummary:"
    assert secho_calls[0][1]["bold"] is True

    assert "Status" in secho_calls[1][0][0]
    assert "Metrics" in secho_calls[1][0][0]

    # --- Rows ---
    expected_rows = [
        ("↑ Improved", 1, "green"),
        ("↓ Regressed", 1, "red"),
        ("± Unchanged", 0, None),
        ("+ Added.", 2, "green"),
        ("- Removed.", 1, "red"),
    ]

    row_calls = secho_calls[3:]  # skip header lines

    for (label, count, color), (args, kwargs) in zip(expected_rows, row_calls):
        line = args[0]

        assert label in line
        assert f"| {count} |" in line

        if color:
            assert kwargs.get("fg") == color
        else:
            assert kwargs.get("dim") is True


def test_print_summary_calls_summarize(
    monkeypatch: pytest.MonkeyPatch, fake_metrics: List
) -> None:
    called = False

    def fake_summarize(metrics: List) -> Dict:
        nonlocal called
        called = True
        return {
            "improved": [],
            "regressed": [],
            "unchanged": [],
            "added": [],
            "removed": [],
        }

    monkeypatch.setattr(
        "rakam_systems_cli.utils.print.summarize",
        fake_summarize,
    )
    monkeypatch.setattr(
        "rakam_systems_cli.utils.print._fmt",
        lambda _: "-",
    )
    monkeypatch.setattr(
        "rakam_systems_cli.utils.print.secho",
        lambda *_, **__: None,
    )

    print_summary(fake_metrics)

    assert called is True



from typing import Any, List, Optional, Tuple


def extract_metric_names(config: Any) -> List[Tuple[str, Optional[str]]]:
    """
    Returns [(type, name)] from EvalConfig / SchemaEvalConfig
    """
    if not hasattr(config, "metrics"):
        return []

    results: List[Tuple[str, Optional[str]]] = []

    for metric in config.metrics or []:
        metric_type = getattr(metric, "type", None)
        metric_name = getattr(metric, "name", None)
        if metric_type:
            results.append((metric_type, metric_name))

    return results



from typing import List
from rakam_systems_cli.cli import app
from typer.testing import CliRunner
from unittest.mock import Mock
from pathlib import Path
from types import SimpleNamespace

import pytest
import typer
from pytest import CaptureFixture

from rakam_systems_cli.cli import metrics


def test_metrics_no_metrics_found(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    file = tmp_path / "eval.py"
    file.write_text("def foo(): pass")

    monkeypatch.setattr(
        "rakam_systems_cli.cli.find_decorated_functions",
        lambda *_: [],
    )

    with pytest.raises(typer.Exit) as exc:
        metrics(directory=tmp_path, recursive=False)

    assert exc.value.exit_code == 0


def test_metrics_finds_metrics(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: CaptureFixture[str],
) -> None:
    file = tmp_path / "eval.py"
    file.write_text("")

    monkeypatch.setattr(
        "rakam_systems_cli.cli.find_decorated_functions",
        lambda *_: ["run_eval"],
    )

    fake_module = SimpleNamespace(
        run_eval=lambda: SimpleNamespace(
            metrics=[
                SimpleNamespace(type="accuracy", name=None),
                SimpleNamespace(type="f1", name="f1_v2"),
            ]
        )
    )

    monkeypatch.setattr(
        "rakam_systems_cli.cli.load_module_from_path",
        lambda _: fake_module,
    )

    metrics(directory=tmp_path, recursive=False)

    out = capsys.readouterr().out

    assert "accuracy" in out
    assert "f1" in out
    assert "unique metrics found" in out


runner = CliRunner()


def test_compare_summary_and_side_by_side_conflict() -> None:
    result = runner.invoke(
        app,
        ["eval", "compare", "--summary", "--side-by-side", "--id", "1", "--id", "2"],
    )

    assert result.exit_code == 1
    assert "--summary and --side-by-side cannot be used together" in result.output


@pytest.mark.parametrize(
    "args",
    [
        ["eval", "compare", "--id", "1"],  # one target
        ["eval", "compare", "--id", "1", "--id",
            "2", "--id", "3"],  # three targets
        ["eval", "compare"],  # none
    ],
)
def test_compare_requires_exactly_two_targets(args: List[str]) -> None:
    result = runner.invoke(app, args)

    assert result.exit_code == 1
    assert "Provide exactly two targets" in result.output


def test_compare_summary_request_failure(monkeypatch: pytest.MonkeyPatch) -> None:
    fake_client = Mock()
    fake_client.compare_testcases.side_effect = RuntimeError("boom")

    monkeypatch.setattr(
        "rakam_systems_cli.cli.DeepEvalClient",
        lambda: fake_client,
    )

    result = runner.invoke(
        app,
        ["eval", "compare", "--summary", "--id", "1", "--id", "2"],
    )

    assert result.exit_code == 1
    assert "Request failed" in result.output


def test_compare_full_mode(monkeypatch: pytest.MonkeyPatch) -> None:
    fake_client = Mock()

    monkeypatch.setattr(
        "rakam_systems_cli.cli.DeepEvalClient",
        lambda: fake_client,
    )

    monkeypatch.setattr(
        "rakam_systems_cli.cli.fetch_run",
        lambda *_, run_id=None, tag=None, **__: (
            {"foo": "a"} if run_id == 1 else {"foo": "b"},
            run_id or tag,
        ),
    )

    serialize_mock = Mock(side_effect=lambda x: f"{x}")
    monkeypatch.setattr(
        "rakam_systems_cli.cli.serialize_for_diff",
        serialize_mock,
    )

    git_diff_mock = Mock()
    monkeypatch.setattr(
        "rakam_systems_cli.cli.git_diff",
        git_diff_mock,
    )

    result = runner.invoke(
        app,
        ["eval", "compare", "--id", "1", "--id", "2"],
    )

    assert result.exit_code == 0

    git_diff_mock.assert_called_once()


def test_compare_fetch_failure(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "rakam_systems_cli.cli.fetch_run",
        Mock(side_effect=RuntimeError("fetch failed")),
    )

    monkeypatch.setattr(
        "rakam_systems_cli.cli.DeepEvalClient",
        lambda: Mock(),
    )

    result = runner.invoke(
        app,
        ["eval", "compare", "--id", "1", "--id", "2"],
    )

    assert result.exit_code == 1
    assert "Fetch failed" in result.output


def test_compare_summary_mode(monkeypatch: pytest.MonkeyPatch) -> None:
    fake_metrics = [SimpleNamespace(
        metric="accuracy", status="added", delta=None)]
    fake_resp = {"metrics": fake_metrics}

    fake_client = Mock()
    fake_client.compare_testcases.return_value = fake_resp

    monkeypatch.setattr(
        "rakam_systems_cli.cli.DeepEvalClient",
        lambda: fake_client,
    )

    # 🔑 Bypass Pydantic validation entirely
    fake_comparison = SimpleNamespace(metrics=fake_metrics)
    monkeypatch.setattr(
        "rakam_systems_cli.cli.TestCaseComparison",
        lambda **_: fake_comparison,
    )

    pretty_mock = Mock()
    monkeypatch.setattr(
        "rakam_systems_cli.cli.pretty_print_comparison",
        pretty_mock,
    )

    result = runner.invoke(
        app,
        ["eval", "compare", "--summary", "--id", "1", "--tag", "baseline"],
    )

    assert result.exit_code == 0

    fake_client.compare_testcases.assert_called_once_with(
        testcase_a_id=1,
        testcase_b_tag="baseline",
        raise_exception=True,
    )

    pretty_mock.assert_called_once_with(
        fake_comparison,
        summary_only=True,
    )



from pathlib import Path
from types import ModuleType

import pytest

from rakam_systems_cli.utils.decorator_utils import load_module_from_path


def test_load_module_from_path_success(tmp_path: Path) -> None:
    file = tmp_path / "mod.py"
    file.write_text(
        """
x = 42
def hello():
    return "world"
"""
    )

    module: ModuleType = load_module_from_path(file)

    assert module.x == 42
    assert module.hello() == "world"


def test_load_module_from_path_invalid(tmp_path: Path) -> None:
    file: Path = tmp_path / "broken.py"
    file.write_text("def foo(")  # invalid syntax

    with pytest.raises(SyntaxError):
        load_module_from_path(file)



import json
from pathlib import Path
from pprint import pprint
from typing import Any, Dict, List, Optional

import typer
from click import Context
from rakam_systems_tools.evaluation.schema import MetricDiff
from typer import secho
from typer.core import TyperGroup


def _print_and_save(
    resp: dict,
    pretty: bool,
    out: Optional[Path],
    overwrite: bool,
) -> None:
    if pretty:
        typer.echo(typer.style("📊 Result:", bold=True))
        pprint(resp)
    else:
        typer.echo(resp)

    if out is None:
        return

    if out.exists() and not overwrite:
        typer.echo(
            f"❌ File already exists: {out} (use --overwrite to replace)")
        raise typer.Exit(code=1)

    out.parent.mkdir(parents=True, exist_ok=True)

    with out.open("w", encoding="utf-8") as f:
        json.dump(resp, f, indent=2, ensure_ascii=False)

    typer.echo(f"💾 Result saved to {out}")


def pct_change(a: Optional[float], b: Optional[float]) -> Optional[str]:
    if a is None or b is None or a == 0:
        return None
    return f"{((b - a) / a) * 100:+.2f}%"


def metric_direction(delta: Optional[float]) -> str:
    if delta is None:
        return "unchanged"
    if delta > 0:
        return "improved"
    if delta < 0:
        return "regressed"
    return "unchanged"


def print_metric_diff(diff: MetricDiff) -> None:
    secho(f"\nMetric: {diff.metric}", bold=True)

    if diff.status == "added":
        secho(f"+ score: {diff.score_b}", fg="green")
        secho(f"+ threshold: {diff.threshold_b}", fg="green")
        secho(f"+ success: {diff.success_b}", fg="green")
        return

    if diff.status == "removed":
        secho(f"- score: {diff.score_a}", fg="red")
        secho(f"- threshold: {diff.threshold_a}", fg="red")
        secho(f"- success: {diff.success_a}", fg="red")
        return

    # unchanged / changed
    if diff.score_a != diff.score_b:
        direction = metric_direction(diff.delta)
        color = "green" if direction == "improved" else "red"
        pct = pct_change(diff.score_a, diff.score_b)

        secho(f"- score: {diff.score_a}", fg="red")
        secho(
            f"+ score: {diff.score_b}" + (f"   ({pct})" if pct else ""),
            fg=color,
        )
    else:
        secho(f"  score: {diff.score_a}", dim=True)

    if diff.threshold_a != diff.threshold_b:
        secho(f"- threshold: {diff.threshold_a}", fg="red")
        secho(f"+ threshold: {diff.threshold_b}", fg="green")
    else:
        secho(f"  threshold: {diff.threshold_a}", dim=True)

    if diff.success_a != diff.success_b:
        secho(f"- success: {diff.success_a}", fg="red")
        secho(f"+ success: {diff.success_b}", fg="green")
    else:
        secho(f"  success: {diff.success_a}", dim=True)


def summarize(metrics: List[MetricDiff]) -> Dict[str, List[str]]:
    """
    Returns metric names grouped by category.
    """
    summary: Dict[str, List[str]] = {
        "improved": [],
        "regressed": [],
        "unchanged": [],
        "added": [],
        "removed": [],
    }

    for m in metrics:
        if m.status == "added":
            summary["added"].append(m.metric)
            continue

        if m.status == "removed":
            summary["removed"].append(m.metric)
            continue

        direction = metric_direction(m.delta)
        summary[direction].append(m.metric)

    return summary


def _fmt(metrics: List[str]) -> str:
    return ", ".join(metrics) if metrics else "-"


def print_summary(metrics: List[MetricDiff]) -> None:
    summary = summarize(metrics)

    rows = [
        ("↑ Improved", "improved", "green"),
        ("↓ Regressed", "regressed", "red"),
        ("± Unchanged", "unchanged", None),
        ("+ Added.", "added", "green"),
        ("- Removed.", "removed", "red"),
    ]

    secho("\nSummary:", bold=True)
    secho(
        "  | Status       | # | Metrics                |",
        dim=True,
    )
    secho(
        "  |--------------|---|------------------------|",
        dim=True,
    )

    for label, key, color in rows:
        count = len(summary[key])
        metrics_str = _fmt(summary[key])

        line = f"  | {label:<12} | {count:<1} | {metrics_str:<22} |"

        if color:
            secho(line, fg=color)
        else:
            secho(line, dim=True)


def pretty_print_comparison(resp: Any, summary_only: bool = False) -> None:
    if not summary_only:
        for metric in resp.metrics:
            print_metric_diff(metric)
        return

    print_summary(resp.metrics)


def serialize_for_diff(obj: dict) -> str:
    """
    Stable, git-friendly JSON representation
    """
    return (
        json.dumps(
            obj,
            indent=4,
            sort_keys=True,
            ensure_ascii=False,
        )
        + "\n"
    )


def git_diff(
    a_text: str,
    b_text: str,
    *,
    label_a: str,
    label_b: str,
    side_by_side: bool = False,
) -> None:
    """
    Show diff between two text blobs.

    - side_by_side: uses git difftool with vimdiff (interactive)
    - fallback to git diff -U3 or difflib if git is not available
    """
    import shutil
    import subprocess
    from pathlib import Path
    from tempfile import TemporaryDirectory

    import typer

    git = shutil.which("git")
    vimdiff = shutil.which("vimdiff")

    with TemporaryDirectory() as tmp:
        a = Path(tmp) / label_a
        b = Path(tmp) / label_b

        a.write_text(a_text)
        b.write_text(b_text)

        # --- Side-by-side with vimdiff ---
        if side_by_side:
            if not git:
                typer.secho(
                    "❌ Git is required for side-by-side diffs", fg="red", bold=True
                )
                return
            if not vimdiff:
                typer.secho(
                    "❌ Vimdiff is not installed. Please install vim or vimdiff to use side-by-side mode.",
                    fg="red",
                    bold=True,
                )
                return

            cmd = [
                git,
                "difftool",
                "--no-index",
                "--tool=vimdiff",
                "--no-prompt",  # skip Y/n prompt
                str(a),
                str(b),
            ]

            subprocess.run(cmd, check=False)
            return

        # (default)
        if git:
            cmd = [
                git,
                "diff",
                "--no-index",
                "--color=always",
                "-U3",
                str(a),
                str(b),
            ]
            subprocess.run(cmd, check=False)
            return

        #  Fallback
        import difflib

        diff = difflib.unified_diff(
            a_text.splitlines(),
            b_text.splitlines(),
            fromfile=label_a,
            tofile=label_b,
            lineterm="",
        )
        for line in diff:
            typer.echo(line)


class OrderedHelpGroup(TyperGroup):
    def format_help(self, ctx: Context, formatter) -> None:
        from click.core import Group
        from click.formatting import HelpFormatter
        from rich.console import Console
        from typer import rich_utils

        console = rich_utils._get_rich_console()
        term_width = console.width

        usage = ctx.command.get_usage(ctx)

        from rich.panel import Panel
        from rich.text import Text

        sections = []

        sections.append(Panel(
            Text(usage, style="bold"),
            title="Usage",
            border_style="cyan",
            width=term_width,
        ))

        cmd_formatter = HelpFormatter(width=term_width, max_width=term_width)
        self.format_commands(ctx, cmd_formatter)
        commands_output = cmd_formatter.getvalue().strip()

        if commands_output:
            lines = [l.strip() for l in commands_output.replace(
                "Commands:", "").strip().split("\n") if l.strip()]
            commands_output = "\n".join(lines)
            sections.append(Panel(
                Text(commands_output),
                title="Commands",
                border_style="green",
                width=term_width,
            ))

        opt_formatter = HelpFormatter(width=term_width, max_width=term_width)
        Group.format_options(self, ctx, opt_formatter)
        options_output = opt_formatter.getvalue().strip()

        if options_output:
            lines = []
            for line in options_output.replace("Options:", "").strip().split("\n"):
                stripped = line.strip()
                if stripped and (stripped.startswith("--") or stripped.startswith("-")):
                    lines.append(stripped)
            options_output = "\n".join(lines)
            if options_output:
                sections.append(Panel(
                    Text(options_output),
                    title="Options",
                    border_style="yellow",
                    width=term_width,
                ))

        for section in sections:
            console.print(section)



from types import SimpleNamespace

from rakam_systems_cli.utils.metric import extract_metric_names


def test_extract_metric_names_no_metrics_attr() -> None:
    config = SimpleNamespace()

    result = extract_metric_names(config)

    assert result == []


def test_extract_metric_names_empty_metrics() -> None:
    config = SimpleNamespace(metrics=[])

    result = extract_metric_names(config)

    assert result == []


def test_extract_metric_names_with_types_only() -> None:
    config = SimpleNamespace(
        metrics=[
            SimpleNamespace(type="accuracy"),
            SimpleNamespace(type="latency"),
        ]
    )

    result = extract_metric_names(config)

    assert result == [
        ("accuracy", None),
        ("latency", None),
    ]


def test_extract_metric_names_with_type_and_name() -> None:
    config = SimpleNamespace(
        metrics=[
            SimpleNamespace(type="accuracy", name="acc_v1"),
            SimpleNamespace(type="f1", name=None),
        ]
    )

    result = extract_metric_names(config)

    assert result == [
        ("accuracy", "acc_v1"),
        ("f1", None),
    ]



# cli.py
import io
import json
import os
import sys
import uuid
from datetime import datetime
from gettext import gettext as _
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import typer
from dotenv import load_dotenv
from rakam_systems_tools.evaluation.client import DeepEvalClient
from rakam_systems_tools.evaluation.schema import TestCaseComparison
from rich.console import Console
from rich.panel import Panel
from rich.pretty import Pretty
from typer.rich_utils import *

from rakam_systems_cli.decorators import eval_run
from rakam_systems_cli.utils.decorator_utils import (
    find_decorated_functions,
    load_module_from_path,
)
from rakam_systems_cli.utils.metric import extract_metric_names
from rakam_systems_cli.utils.print import (
    OrderedHelpGroup,
    _print_and_save,
    git_diff,
    pretty_print_comparison,
    serialize_for_diff,
)

console = Console()


load_dotenv()
app = typer.Typer(help="Rakam CLI tools", cls=OrderedHelpGroup)


# add root of the project to sys.path
PROJECT_ROOT = os.path.abspath(".")
if PROJECT_ROOT not in sys.path:
    sys.path.insert(0, PROJECT_ROOT)
list_app = typer.Typer(help="List evaluations or runs", cls=OrderedHelpGroup)
metrics_app = typer.Typer(help="Metrics utilities", cls=OrderedHelpGroup)
eval_app = typer.Typer(help="Evaluation utilities", cls=OrderedHelpGroup)


@metrics_app.command("list")
def metrics(
    directory: Path = typer.Argument(
        Path("./eval"),
        exists=True,
        file_okay=False,
        dir_okay=True,
        help="Directory to scan (default: ./eval)",
    ),
    recursive: bool = typer.Option(
        False,
        "-r",
        "--recursive",
        help="Recursively search for Python files",
    ),
) -> None:
    """
    List all metric types used by loaded eval configs.
    """
    files = directory.rglob("*.py") if recursive else directory.glob("*.py")
    TARGET_DECORATOR = eval_run.__name__

    all_metrics: Set[Tuple[str, Optional[str]]] = set()
    found_any = False

    for file in sorted(files):
        functions = find_decorated_functions(file, TARGET_DECORATOR)
        if not functions:
            continue

        typer.echo(f"\n📄 {file}")

        try:
            module = load_module_from_path(file)
        except Exception as e:
            typer.echo(f"  ❌ Failed to import module: {e}")
            continue

        for fn_name in functions:
            try:
                func = getattr(module, fn_name)
                result = func()

                metrics = extract_metric_names(result)
                if not metrics:
                    continue

                found_any = True
                for metric_type, metric_name in metrics:
                    all_metrics.add((metric_type, metric_name))

                    if metric_name:
                        typer.echo(f"  • {metric_type} (alias: {metric_name})")
                    else:
                        typer.echo(f"  • {metric_type}")

            except Exception as e:
                typer.echo(f"  ❌ Failed to inspect {fn_name}: {e}")

    if not found_any:
        typer.echo("\nNo metrics found.")
        raise typer.Exit(code=0)

    typer.echo(f"\n✅ {len(all_metrics)} unique metrics found")


@list_app.command("evals")
def list_evals(
    directory: Path = typer.Argument(
        Path("./eval"),
        exists=True,
        file_okay=False,
        dir_okay=True,
        help="Directory to scan (default: ./eval)",
    ),
    recursive: bool = typer.Option(
        False,
        "--recursive",
        "-r",
        help="Recursively search for Python files",
    ),
) -> None:
    """
    List evaluations (functions decorated with @eval_run).
    """
    TARGET_DECORATOR = eval_run.__name__
    files = directory.rglob("*.py") if recursive else directory.glob("*.py")

    found = False

    for file in sorted(files):
        functions = find_decorated_functions(file, TARGET_DECORATOR)
        for fn in functions:
            found = True
            typer.echo(f"{file}:{fn}")

    if not found:
        typer.echo("No evaluations found.")


@list_app.command("runs")
def list_runs(
    limit: int = typer.Option(20, "-l", "--limit", help="Max number of runs"),
    offset: int = typer.Option(0, help="Pagination offset"),
) -> None:
    """
    List runs (newest first).
    """
    client = DeepEvalClient()

    response = client.list_evaluation_testcases(
        limit=limit,
        offset=offset,
        raise_exception=True,
    )
    assert response is not None
    items = response.get("items", [])
    total = response.get("total", 0)

    if not items:
        typer.echo("No runs found.")
        return

    typer.echo(f"[id] {'tag':<20}{'label':<20}created_at")

    for run in items:
        run_id = run.get("id")
        label = run.get("label") or "-"
        uid = run.get("tag") or "-"
        created_at = run.get("created_at")

        if created_at:
            try:
                created_at = datetime.fromisoformat(created_at).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
            except ValueError:
                pass

        typer.echo(f"[{run_id}] {uid:<20} {label:<20} {created_at}")

    shown = offset + len(items)
    if shown < total:
        typer.echo()
        typer.echo(
            f"Showing {shown} of {total} runs. Use --limit to see more.")


@eval_app.command()
def show(
    run_id: Optional[int] = typer.Option(
        None,
        "-i",
        "--id",
        help="Run ID",
    ),
    tag: Optional[str] = typer.Option(
        None,
        "-t",
        "--tag",
        help="Run tag",
    ),
    raw: bool = typer.Option(
        False,
        "--raw",
        help="Print raw JSON instead of formatted output",
    ),
) -> None:
    """
    Show a run by ID or tag. Without arguments, shows the most recent run.
    """

    if run_id and tag:
        raise typer.BadParameter("Provide only one of --id or --tag")

    client = DeepEvalClient()

    # If no arguments, fetch the most recent run
    if not run_id and not tag:
        response = client.list_evaluation_testcases(
            limit=1, offset=0, raise_exception=True
        )
        if not response or not response.get("items"):
            console.print(
                Panel(
                    "No runs found",
                    title="Error",
                    style="red",
                )
            )
            raise typer.Exit(code=1)
        run_id = response["items"][0]["id"]
        identifier = f"run_id={run_id} (latest)"
        assert isinstance(run_id, int)
        result = client.get_evaluation_testcase_by_id(run_id)
    elif run_id:
        result = client.get_evaluation_testcase_by_id(run_id)
        identifier = f"run_id={run_id}"
    else:
        assert tag is not None
        result = client.get_evaluation_testcase_by_tag(tag)
        identifier = f"tag={tag}"

    if not result:
        console.print(
            Panel(
                f"No response received for {identifier}",
                title="Error",
                style="red",
            )
        )
        raise typer.Exit(code=1)

    if isinstance(result, dict) and result.get("error"):
        console.print(
            Panel(
                result["error"],
                title="Error",
                style="red",
            )
        )
        raise typer.Exit(code=1)

    if raw:
        console.print(Pretty(result))
        raise typer.Exit()

    console.print(
        Panel.fit(
            Pretty(result),
            title="Run",
            subtitle=identifier,
        )
    )


def validate_eval_result(result: Any, fn_name: str) -> str:
    eval_config = getattr(result, "__eval_config__", None)

    if not isinstance(eval_config, str):
        expected = "EvalConfig or SchemaEvalConfig"
        actual = type(result).__name__

        typer.echo(
            f"    ❌ Invalid return type from `{fn_name}`\n"
            f"       Expected: {expected}\n"
            f"       Got: {actual}"
        )
        return ""

    return eval_config


@eval_app.command()
def run(
    directory: Path = typer.Argument(
        Path("./eval"),
        exists=True,
        file_okay=False,
        dir_okay=True,
        help="Directory to scan (default: ./eval)",
    ),
    recursive: bool = typer.Option(
        False,
        "-r",
        "--recursive",
        help="Recursively search for Python files",
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        help="Only list functions without executing them",
    ),
    save_runs: bool = typer.Option(
        False,
        "--save-runs",
        help="Save each run result to a JSON file",
    ),
    output_dir: Path = typer.Option(
        Path("./eval_runs"),
        "--output-dir",
        help="Directory where run results are saved",
    ),
) -> None:
    """
    Execute evaluations (functions decorated with @eval_run).
    """
    files = directory.rglob("*.py") if recursive else directory.glob("*.py")
    TARGET_DECORATOR = eval_run.__name__

    executed_any = False

    if save_runs and not dry_run:
        output_dir.mkdir(parents=True, exist_ok=True)

    for file in sorted(files):
        functions = find_decorated_functions(file, TARGET_DECORATOR)
        if not functions:
            continue

        typer.echo(f"\n📄 {file}")

        try:
            module = load_module_from_path(file)
        except Exception as e:
            typer.echo(f"  ❌ Failed to import module: {e}")
            continue

        for fn_name in functions:
            typer.echo(f"  ▶ {fn_name}")

            try:
                func = getattr(module, fn_name)
                result = func()

                eval_type = validate_eval_result(result, fn_name)
                if not eval_type:
                    continue

                if dry_run:
                    typer.echo(f"    🧪 Dry-run OK → {eval_type}")
                    continue

                client = DeepEvalClient()

                if eval_type == "text_eval":
                    resp = client.text_eval(config=result)
                else:
                    resp = client.schema_eval(config=result)

                typer.echo(f"{resp}")
                typer.echo(f"    ✅ Returned {type(result).__name__}")

                if save_runs:
                    run_id = (
                        resp["id"]
                        if resp is not None and "id" in resp
                        else uuid.uuid4().hex[:8]
                    )

                    output_path = output_dir / f"run_{fn_name}_{run_id}.json"

                    def to_json_safe(obj: Any) -> Any:
                        if hasattr(obj, "model_dump"):
                            return obj.model_dump()
                        if hasattr(obj, "dict"):
                            return obj.dict()
                        return obj

                    with output_path.open("w", encoding="utf-8") as f:
                        json.dump(
                            to_json_safe(resp),
                            f,
                            indent=2,
                            ensure_ascii=False,
                        )

                    typer.echo(f"    💾 Saved run → {output_path}")

            except Exception as e:
                typer.echo(f"    ❌ Execution failed: {e}")


def fetch_run(
    client: DeepEvalClient,
    *,
    run_id: Optional[int],
    tag: Optional[str],
) -> Tuple[dict, str]:
    """
    Fetch a single run by id or tag.
    Returns (payload, identifier)
    """
    if run_id is not None:
        result = client.get_evaluation_testcase_by_id(run_id)
        identifier = f"run_id={run_id}"
    else:
        assert tag is not None
        result = client.get_evaluation_testcase_by_tag(tag)
        identifier = f"tag={tag}"

    if not result:
        raise RuntimeError(f"No data returned for {identifier}")

    return result, identifier


@eval_app.command()
def compare(
    tag: List[str] = typer.Option(
        [],
        "-t",
        "--tag",
        help="Run tag",
    ),
    run_id: List[int] = typer.Option(
        [],
        "-i",
        "--id",
        help="Run ID",
    ),
    summary: bool = typer.Option(
        False,
        "--summary",
        help="Show summary diff only",
    ),
    side_by_side: bool = typer.Option(
        False,
        "--side-by-side",
        help="Show side-by-side diff (git)",
    ),
) -> None:
    """
    Compare two evaluation runs.

    Default: unified git diff
    """

    if summary and side_by_side:
        typer.secho(
            "❌ --summary and --side-by-side cannot be used together",
            fg="red",
            bold=True,
        )
        raise typer.Exit(code=1)

    targets: List[Tuple[str, Union[str, int]]] = []

    for r in run_id:
        targets.append(("run", r))
    for t in tag:
        targets.append(("tag", t))

    if len(targets) != 2:
        typer.secho(
            "❌ Provide exactly two targets using --id and/or --tag",
            fg="red",
            bold=True,
        )
        raise typer.Exit(code=1)

    client = DeepEvalClient()
    # Summary mode (reduced payload)
    (type_a, value_a), (type_b, value_b) = targets
    if summary:
        kwargs: Dict[str, Any] = {"raise_exception": True}
        if type_a == "run":
            kwargs["testcase_a_id"] = value_a
        else:
            kwargs["testcase_a_tag"] = value_a

        if type_b == "run":
            kwargs["testcase_b_id"] = value_b
        else:
            kwargs["testcase_b_tag"] = value_b
        try:
            resp = client.compare_testcases(**kwargs)
        except Exception as e:
            typer.secho(f"❌ Request failed: {e}", fg="red")
            raise typer.Exit(code=1)

        if not resp:
            typer.secho("⚠️ No response received", fg="yellow")
            raise typer.Exit(code=1)
        comparison = TestCaseComparison(**resp)
        pretty_print_comparison(
            comparison,
            summary_only=summary,
        )
        return

    try:
        run_a, id_a = fetch_run(
            client,
            run_id=int(value_a) if type_a == "run" else None,
            tag=str(value_a) if type_a == "tag" else None,
        )
        run_b, id_b = fetch_run(
            client,
            run_id=int(value_b) if type_b == "run" else None,
            tag=str(value_b) if type_b == "tag" else None,
        )
    except Exception as e:
        typer.secho(f"❌ Fetch failed: {e}", fg="red")
        raise typer.Exit(code=1)

    a_text = serialize_for_diff(run_a)
    b_text = serialize_for_diff(run_b)

    git_diff(
        a_text,
        b_text,
        label_a=f"{id_a}.full.json",
        label_b=f"{id_b}.full.json",
        side_by_side=side_by_side,
    )


@eval_app.command(hidden=True)
def compare_label_latest(
    label_a: str = typer.Argument(
        ...,
        help="First label (latest run will be used)",
    ),
    label_b: str = typer.Argument(
        ...,
        help="Second label (latest run will be used)",
    ),
    pretty: bool = typer.Option(
        True,
        "--pretty/--raw",
        help="Pretty-print the response",
    ),
    raise_exception: bool = typer.Option(
        False,
        "--raise",
        help="Raise HTTP exceptions instead of swallowing them",
    ),
    out: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out",
        help="Optional file path to save the result as JSON",
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        help="Overwrite output file if it already exists",
    ),
) -> None:
    """
    Compare the latest runs for two labels.
    """
    client = DeepEvalClient()

    typer.echo(f"🔍 Comparing latest runs: '{label_a}' ↔ '{label_b}'")

    try:
        resp = client.compare_latest_by_labels(
            label_a=label_a,
            label_b=label_b,
            raise_exception=raise_exception,
        )
    except Exception as e:
        typer.echo(f"❌ Request failed: {e}")
        raise typer.Exit(code=1)

    if not resp:
        typer.echo("⚠️ No response received")
        raise typer.Exit(code=1)

    _print_and_save(resp, pretty, out, overwrite)


@eval_app.command(hidden=True)
def compare_last(
    label: str = typer.Argument(
        ...,
        help="Label whose last two runs will be compared",
    ),
    pretty: bool = typer.Option(
        True,
        "--pretty/--raw",
        help="Pretty-print the response",
    ),
    raise_exception: bool = typer.Option(
        False,
        "--raise",
        help="Raise HTTP exceptions instead of swallowing them",
    ),
    out: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out",
        help="Optional file path to save the result as JSON",
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        help="Overwrite output file if it already exists",
    ),
) -> None:
    """
    Compare the last two evaluation runs of a label.
    """
    client = DeepEvalClient()

    typer.echo(f"🔍 Comparing last two runs for label '{label}'")

    try:
        resp = client.compare_last_two_by_label(
            label=label,
            raise_exception=raise_exception,
        )
    except Exception as e:
        typer.echo(f"❌ Request failed: {e}")
        raise typer.Exit(code=1)

    if not resp:
        typer.echo("⚠️ No response received")
        raise typer.Exit(code=1)

    _print_and_save(resp, pretty, out, overwrite)


@eval_app.command("tag")
def tag_command(
    run_id: Optional[int] = typer.Option(
        None,
        "-i",
        "--id",
        help="Run ID",
    ),
    tag: Optional[str] = typer.Option(
        None,
        "-t",
        "--tag",
        help="Tag to assign to the run",
    ),
    delete: Optional[str] = typer.Option(
        None,
        "--delete",
        help="Delete a tag",
    ),
) -> None:
    """
    Assign a tag to a run or delete a tag.
    """

    if delete:
        if run_id or tag:
            typer.echo("❌ --delete cannot be used with --id or --tag")
            raise typer.Exit(code=1)
    else:
        if not run_id or not tag:
            typer.echo("❌ Use --id and --tag together, or --delete")
            raise typer.Exit(code=1)

    client = DeepEvalClient()

    if delete:
        assert run_id is not None

        result = client.update_evaluation_testcase_tag(
            testcase_id=run_id,
            tag=delete,
            raise_exception=True,
        )
        typer.echo("🗑️ Tag deleted successfully")
        typer.echo(f"Tag: {delete}")
        return
    assert run_id is not None
    assert tag is not None
    result = client.update_evaluation_testcase_tag(
        testcase_id=run_id,
        tag=tag,
        raise_exception=True,
    )
    assert result is not None

    typer.echo("✅ Tag assigned successfully")
    typer.echo(f"Run ID: {run_id}")
    typer.echo(f"Tag: {result.get('tag')}")


# Register sub-apps in user journey order (after regular commands)
eval_app.add_typer(list_app, name="list")
eval_app.add_typer(metrics_app, name="metrics")
app.add_typer(eval_app, name="eval")


def main() -> None:
    app()


if __name__ == "__main__":
    main()
